options { language = Cpp;
}
class janlex extends Lexer;
options{
  	k=2;
    genHashLines = true;		// include line number information
}
tokens{INTEGER; READ_DENOTATION;}

   IDENTIFIER : ('a'..'z'|'A'..'Z') ('a'..'z'|'A'..'Z'|'0'..'9'|'_')*;
   NUMBER
      : ('0'..'9')+
        ( ('E'|'B') ('-')? ('0'..'9')+ { $setType(REAL_DENOTATION); }
        | { $setType(INTEGER); }
        )
      ;
/*
   INTEGER : ('0'..'9')+;
   REAL_DENOTATION : ('0'..'9')+ ('E'|'B') ('-')? INTEGER;
*/
   STRING :  '\"' (~'\"' | '\"' '\"')* '\"';
   EOL : '.'(~'\n')*'\n';
   X : ' ';
//   EOF


class janparse extends Parser;
  program : module;

  module:
     "START" X IDENTIFIER EOL 
        (specification_element EOL)+ 
        "FINISH" EOL 
        EOF; 
  symbol : IDENTIFIER;

  specification_element: 
     manifest_definition | 
     global_definition | 
     range_definition | 
     precision_definition | 
     array_mode_definition | 
     record_mode_definition | 
     space_declaration | 
     common_declaration | 
     procedure_declaration; 
  manifest_definition:
     "DEF" X IDENTIFIER X elementary_value; 
  expression:
     ("-")? term (("+" | "-") (term ( ("+" | "-") term)* ))?; 
  term: primary ( ("*" | "/") primary)*; 
  primary: INTEGER | IDENTIFIER | "(" expression ")"; 
  global_identifier: IDENTIFIER;

  mode: IDENTIFIER; 
  pre_defined_mode: IDENTIFIER;
  int_mode: IDENTIFIER;
  real_mode: IDENTIFIER;
  range_definition:
     "RANGE" X symbol X expression X expression; 
  precision_definition:
     "PRECSN" X symbol X expression X (expression); 
  denotation:
     program_storage X IDENTIFIER | 
     code_location | 
     "NIL" | 
    "FALSE" | "TRUE" | 
     character_denotation | 
     ("-")? (INTEGER | REAL_DENOTATION);

  character_denotation : STRING; // TODO: limit to one character
  pack: "PACK" | "ALIGN";
  element_count:
     "(" expression ")" | "[" expression "]"; 
  array_mode_definition:
     "ARRAY" X symbol X pack X mode element_count; 
 record_mode_definition:
     "RECORD" X symbol EOL 
        record_layout 
        "RECEND" X symbol; 
 record_layout
     : 
       ("FIELD" X mode X pack X symbol EOL)+ 
       ( "MARK" X INTEGER X variant_id EOL (variant)+)?
     | "MARK" X INTEGER X variant_id EOL (variant)+
     ; 

 variant: 
     record_layout 
       "BACK" X INTEGER X variant_id EOL; 
  variant_id: symbol; 
  composite_denotation
     : ("ELMV" X mode X valuable X element_count (X value)? EOL)+ 
     | ("FLDV" X mode X valuable X selector (X value)? EOL)+
     ;
  valuable: "LIT" | "CONST" | program_storage; 
  selector: IDENTIFIER;

  value:
     elementary_value | composite_value;

  elementary_value:
     "A" X mode X denotation | 
     ("N" | "T") X real_mode X ("-")? REAL_DENOTATION | 
     "C" X int_mode X character_denotation | 
     "E" X int_mode X expression | 
     "M" X pre_defined_mode X IDENTIFIER; 
  code_location:
     ("R" | "X" | "Y") X label;
  composite_value:
     "V" X mode EOL 
        composite_denotation 
        "VEND" X mode X valuable X (mode);

  category:
     program_storage | 
     module_storage | 
     contour_storage | 
     parameter_storage | 
     temporary_storage | 
     "CONST"; 
  program_storage: "STATIC" ("(" symbol ")")?; 
  module_storage: "INTERN" ("(" symbol ")")?; 
  contour_storage: "DISP" ("(" INTEGER ")")?; 
  parameter_storage: "PARAM" ("(" INTEGER ")")?; 
  temporary_storage: "TEMP" ("(" INTEGER ")")?;

  reference:
     (IDENTIFIER ( index )? | index ) (X selector)?; 
  index:
     ("(" expression ")" | "[" expression "]") 
        mode ("+" | "*")?; 
  constant: "CONST" X reference;
  variable:
     (program_storage | module_storage | contour_storage | 
        parameter_storage) X reference | 
     "BASED" X (reference); 
  temporary:
     temporary_storage X reference;

  space_declaration:
     "SPACE" X mode X category X symbol (X value)?;

  common_declaration:
     "COMMON" X mode X category X (global_identifier (X value)? )?;

  block_declaration:
     "BLOCK" X INTEGER X procedure_class EOL 
        (procedure_element EOL)+ 
        "BLEND" X INTEGER X procedure_class;

  release_declaration:
     "RELEASE" X mode X assignable X IDENTIFIER; 
  assignable:
     program_storage | 
     module_storage | 
     contour_storage | 
     parameter_storage | 
     temporary_storage;
  data_location: constant | variable;
  cell: temporary | variable;

  procedure_declaration:
     "BEGIN" X procedure_heading EOL 
        ("PARAM" X mode X procedure_class X symbol EOL)* 
        "PAREND" X procedure_heading EOL 
        (specification_element EOL)* 
        "BODY" X procedure_heading EOL 
        (procedure_element EOL)+ 
        "END" X procedure_heading; 
  procedure_heading: 
     (mode)? X procedure_class X label; 
  procedure_class: "NONREC" | "MAIN" | "REC" "(" INTEGER ")"; 
  label: symbol; 
  procedure_element: 
     manifest_definition | 
     global_definition | 
     range_definition | 
     precision_definition | 
     array_mode_definition | 
     record_mode_definition | 
     space_declaration | 
     common_declaration | 
     block_declaration | 
     release_declaration | 
     label_declaration | 
     imperative_statement;

  label_declaration: "LOC" X label;
  global_definition:
     ("GLOBAL" | "FOREIGN") X global_identifier X 
        ("ENTRY" | "EXTERN" | "SYSTEM") X label;

  imperative_statement:
     transmission | 
     computation | 
     control_statement | 
     allocation | 
     test_panel_instruction;

  operand:
     value | constant | temporary | variable | size; 
  size: "SIZE" index;
  transmission:
     load_operation | 
     move_operation | 
     base_index_operation | 
     vector_operation;

  load_operation:
     "LOAD" X mode X operand | 
     "SETLOC" X mode X data_location;

  move_operation:
     "STORE" ("(N)")? X mode X cell | 
     "MOVE" X mode X operand EOL 
        (base_index_operation EOL)* 
        "TO" X mode X cell;

  base_index_operation:
     "BASE" ("(R)")* X mode X (operand) | 
     "BASLOC" X mode X (constant | cell) | 
     "INDEX" ("(R)")? X mode X (operand);

  vector_operation:
     "VECTOR" X mode X operatorr X modifier EOL 
        ((base_index_operation EOL)* 
           ("SOURCE" X mode X operand | 
           "TARGET" X mode X variable | 
           "COUNT" X int_mode X (operand)) EOL)+ 
        "VECEND" X mode X operatorr X modifier;

  computation:
     operatorr X modifier X (operand); 
  operatorr: IDENTIFIER ("(R)" | "(N)")?; 
  modifier: IDENTIFIER;

  control_statement:
     "JMP" X (condition_code) X (target) | 
     "TRAP" X (condition_code) X IDENTIFIER | 
     "BEGCAS" X int_mode X (operand) EOL 
        ("CASE" X int_mode X 
           code_location X elementary_value EOL)+ 
        "ENDCAS" X int_mode X (code_location) | 
     call | 
     "PROC" X contour_storage X (target) | 
     "RETURN" X procedure_heading | 
     "STOP" X procedure_heading | 
     "EXIT" X (condition_code) X (cell) | 
     "ABORT" X procedure_heading; 
  target: 
     code_location | constant | temporary | variable;

  condition_code: ;
  call:
     "RCALL" X procedure_operand EOL 
        (recursive_argument EOL)* 
        "RCEND" X procedure_operand | 
     "CALL" X procedure_operand EOL 
        (nonrecursive_argument EOL)* 
        "CEND" X procedure_operand; 
  procedure_operand: (mode) X (target); 
  recursive_argument: 
     (procedure_element EOL)* 
     ("RARG" X mode X (operand) | 
        "RARGLOC" X mode X data_location); 
  nonrecursive_argument: 
     (procedure_element EOL)* 
     ("ARG" X mode X (operand) | 
        "ARGLOC" X mode X data_location);

  allocation:
     ("GRAB" | "FREE") X ("DISP" | "HEAP") X size;

  test_panel_instruction:
     "PR" "=" STRING | 
     "LOOK" ("(N)")? X mode X (operand) | 
     "CC" "=" condition_code;
